/* -*-c-*-
 *
 * $Id$
 *
 * Various useful utilities, stolen from mLib
 *
 * (c) 2001 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Anag: a simple wordgame helper.
 *
 * Anag is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Anag is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Anag; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "anag.h"

/*----- Static variables --------------------------------------------------*/

static const char *quis = "<unset>";

/*----- Error reporting ---------------------------------------------------*/

/* --- @ego@ --- *
 *
 * Arguments:	@const char *p@ = pointer to program name
 *
 * Returns:	---
 *
 * Use:		Stores what the program's name is.
 */

#ifndef PATHSEP
#  if defined(__riscos)
#    define PATHSEP '.'
#  elif defined(__unix) || defined(unix)
#    define PATHSEP '/'
#  else
#    define PATHSEP '\\'
#  endif
#endif

void ego(const char *p)
{
  const char *q = p;
  while (*q) {
    if (*q++ == PATHSEP)
      p = q;
  }
  if (*p == '-')
    p++;
  quis = p;
}

#undef PATHSEP

/* --- @pquis@ --- *
 *
 * Arguments:	@FILE *fp@ = output stream to write on
 *		@const char *p@ = pointer to string to write
 *
 * Returns:	Zero if everything worked, EOF if not.
 *
 * Use:		Writes the string @p@ to the output stream @fp@.  Occurrences
 *		of the character `$' in @p@ are replaced by the program name
 *		as reported by @quis@.  A `$$' is replaced by a single `$'
 *		sign.
 */

int pquis(FILE *fp, const char *p)
{
  size_t sz;

  while (*p) {
    sz = strcspn(p, "$");
    if (sz) {
      if (fwrite(p, 1, sz, fp) < sz)
	return (EOF);
      p += sz;
    }
    if (*p == '$') {
      p++;
      if (*p == '$') {
	if (fputc('$', fp) == EOF)
	  return (EOF);
	p++;
      } else {
	if (fputs(quis, fp) == EOF)
	  return (EOF);
      }
    }
  }
  return (0);
}

/* --- @die@ --- *
 *
 * Arguments:	@const char *f@ = a @printf@-style format string
 *		@...@ = other arguments
 *
 * Returns:	Never.
 *
 * Use:		Reports an error and exits.
 */

void die(const char *f, ...)
{
  va_list ap;
  va_start(ap, f);
  fprintf(stderr, "%s: ", quis);
  vfprintf(stderr, f, ap);
  va_end(ap);
  putc('\n', stderr);
  exit(EX_FAIL);
}

/*----- Memory allocation -------------------------------------------------*/

/* --- @xmalloc@ --- *
 *
 * Arguments:	@size_t sz@ = size of block to allocate
 *
 * Returns:	Pointer to allocated block.
 *
 * Use:		Allocates memory.  If there's not enough memory, the
 *		program exits.
 */

void *xmalloc(size_t sz)
{
  void *p = malloc(sz);
  if (!p)
    die("not enough memory");
  return (p);
}

/* --- @xrealloc@ --- *
 *
 * Arguments:	@void *p@ = a pointer to allocated memory
 *		@size_t sz@ = new size of block wanted
 *
 * Returns:	Pointer to resized block.
 *
 * Use:		Resizes an allocated block.  If there's not enough memory,
 *		the program exits.
 */

void *xrealloc(void *p, size_t sz)
{
  p = realloc(p, sz);
  if (!p)
    die("not enough memory");
  return (p);
}

/*----- Dynamic string handling -------------------------------------------*/

#define DSTR_INITSZ 64

/* --- @dstr_destroy@ --- *
 *
 * Arguments:	@dstr *d@ = pointer to a dynamic string block
 *
 * Returns:	---
 *
 * Use:		Reclaims the space used by a dynamic string.
 */

void dstr_destroy(dstr *d) { free(d->buf); d->len = 0; d->sz = 0; }

/* --- @dstr_reset@ --- *
 *
 * Arguments:	@dstr *d@ = pointer to a dynamic string block
 *
 * Returns:	---
 *
 * Use:		Resets a string so that new data gets put at the beginning.
 */

void dstr_reset(dstr *d) { d->len = 0; }

/* --- @dstr_ensure@ --- *
 *
 * Arguments:	@dstr *d@ = pointer to a dynamic string block
 *		@size_t sz@ = amount of free space to ensure
 *
 * Returns:	---
 *
 * Use:		Ensures that at least @sz@ bytes are available in the
 *		given string.
 */

void dstr_ensure(dstr *d, size_t sz)
{
  size_t rq = d->len + sz;
  size_t nsz;

  /* --- If we have enough space, just leave it --- */

  if (rq <= d->sz)
    return;

  /* --- Grow the buffer --- */

  nsz = d->sz;

  if (nsz == 0)
    nsz = (DSTR_INITSZ >> 1);
  do nsz <<= 1; while (nsz < rq);

  if (d->buf)
    d->buf = xrealloc(d->buf, nsz);
  else
    d->buf = xmalloc(nsz);
  d->sz = nsz;
}

/* --- @dstr_putline@ --- *
 *
 * Arguments:	@dstr *d@ = pointer to a dynamic string block
 *		@FILE *fp@ = a stream to read from
 *
 * Returns:	The number of characters read into the buffer, or @EOF@ if
 *		end-of-file was reached before any characters were read.
 *
 * Use:		Appends the next line from the given input stream to the
 *		string.  A trailing newline is not added; a trailing null
 *		byte is appended, as for @dstr_putz@.
 */

int dstr_putline(dstr *d, FILE *fp)
{
  size_t left = d->sz - d->len;
  size_t off = d->len;
  int rd = 0;
  int ch;

  for (;;) {

    /* --- Read the next byte --- */

    ch = getc(fp);

    /* --- End-of-file when no characters read is special --- */

    if (ch == EOF && !rd)
      return (EOF);

    /* --- Make sure there's some buffer space --- */

    if (!left) {
      d->len = off;
      dstr_ensure(d, 1);
      left = d->sz - off;
    }

    /* --- End-of-file or newline ends the loop --- */

    if (ch == EOF || ch == '\n') {
      d->buf[off] = 0;
      d->len = off;
      return rd;
    }

    /* --- Append the character and continue --- */

    d->buf[off++] = ch;
    left--; rd++;
  }
}

/*----- That's all, folks -------------------------------------------------*/
